/*
    Serial port logic analyzer suite
    Copyright 1999 Terran Melconian (terran@consistent.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program, in the file "LICENSE".  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
    Boston, MA  02111-1307  USA

*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/io.h>
#include <string.h>
#include <values.h>

#define TRANSITIONS 100000
int maxidenticallines=0;

union MSR /* Modem Status Register */
{
    char byte;
    struct
    {
	int DCTS:1;
	int DDSR:1;
	int TERI:1;
	int DDCD:1;
	int CTS:1;
	int DSR:1;
	int RI:1;
	int DCD:1;
    } flags;
};

struct TRANSITION /* Recording signal level transitions */
{
    int tick; /* How many clock ticks in from beginning */
    long loop; /* and how many loops since that tick */
    union MSR value; /* Record the new values */
};

struct TRANSITION transitions[TRANSITIONS];

void chart(long start, long stop, long usecdiv)
{
    long counter, loop;
    union MSR state;
    int alias;
    char buf[256]="";

    /* Used to remove redundant lines */
    char identicalbyte=1;
    int identicalcount=0;

    /* Read in the data from the sample module output */
    
    for (counter=0; counter<TRANSITIONS; counter++)
    {
	if (scanf("%d %ld %x\n", &transitions[counter].tick,
	    &transitions[counter].loop, &transitions[counter].value.byte)
	    != 3)
	    break;
	if (transitions[counter].tick==0)
	    break;
    }

    /* Advance the start time to the first data point, approximately */
    if (start < transitions[0].loop - usecdiv*10)
	start = transitions[0].loop - usecdiv*10;
    
    /* Now just print the chart */
    
    printf ("    uSec    CTS      DSR      RI       DCD\n\n");
    
#define HIGH "  .   |  "
#define LOW  "  |   .  "
    
    loop=0;
    for (counter=start; counter<=stop; counter+=usecdiv)
    {
	/* If there has been a transition */
	alias=0;
	while (transitions[loop].loop <= counter &&
		transitions[loop].tick)
	{
	    if (state.byte != transitions[loop].value.byte)
		alias++;
	    state.byte=transitions[loop].value.byte;
	    loop++;
	}
	
	if (alias > 1 && counter != start)
	    fprintf (stderr, "Warning!  You are getting aliasing.  "
		    "Try smaller timesstep.\n");
	
	/* If there's no more data, stop */
	if (transitions[loop].tick == 0)
	    break;

	if (identicalbyte == state.byte) {
	    identicalcount++;
	} else {
	    /* Print last identical line, for context, if we were suppressing */
	    if (identicalcount >= maxidenticallines && maxidenticallines)
		fputs(buf, stdout);

	    identicalbyte = state.byte;
	    identicalcount=0;
	}

	sprintf (buf, "%9ld%9s%9s%9s%9s\n", counter,
		(state.flags.CTS)?HIGH:LOW,
		(state.flags.DSR)?HIGH:LOW,
		(state.flags.RI)?HIGH:LOW,
		(state.flags.DCD)?HIGH:LOW);

	/* Suppress printing too many identical lines in a row */
	if (identicalcount == maxidenticallines && maxidenticallines > 0)
	    printf("%9s\n", "...");
	if (identicalcount >= maxidenticallines && maxidenticallines > 0)
	    continue;
	else
	    fputs(buf, stdout);
    }
}

int main (int argc, char*argv[])
{
    int i;
    long usecdiv=10;
    long start=0,stop=MAXLONG;
    
    for (i=1; i<argc; i++)
    {
	if (strcmp(argv[i], "-u")==0)
	    usecdiv=atol(argv[++i]);
	else if (strcmp(argv[i], "-i")==0)
	{
	    start=atol(argv[++i])*1000;
	    stop=atol(argv[++i])*1000;
	}
	else if (strcmp(argv[i], "-I")==0)
	{
	    start=atol(argv[++i]);
	    stop=atol(argv[++i]);
	}
	else if (strcmp(argv[i], "-f")==0)
	{
	    maxidenticallines=atoi(argv[++i]);
	}
	else
	{
	    printf ("Valid options are:\n\n"
		     "  -u n    display n usec per line\n"
		     "  -f n    suppress more than n repeated lines\n"
		     "  -i n m  display interval from n to m msec (defaults to all)\n"
		     "  -I n m  display interval from n to m usec\n\n");
	    exit(255);
	}
    }
    
    chart(start,stop,usecdiv);

    return 0;
}
