/*
    Serial port logic analyzer suite
    Copyright 1999 Terran Melconian (terran@consistent.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program, in the file "LICENSE".  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
    Boston, MA  02111-1307  USA

*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/io.h>
#include <string.h>
#include <values.h>

/* This is data dumper */

#define TRANSITIONS 100000
#define MAXWORDSIZE 100


union MSR /* Modem Status Register */
{
    char byte;
    struct
    {
	int DCTS:1;
	int DDSR:1;
	int TERI:1;
	int DDCD:1;
	int CTS:1;
	int DSR:1;
	int RI:1;
	int DCD:1;
    } flags;
};

struct TRANSITION /* Recording signal level transitions */
{
    int tick; /* How many clock ticks in from beginning */
    long loop; /* and how many loops since that tick */
    union MSR value; /* Record the new values */
};

struct TRANSITION transitions[TRANSITIONS];

struct RSTRANS /* Modified version for use here */
{
    unsigned long time; /* time in microseconds */
    char state; /* New state, 1 or 0 */
};
struct RSTRANS rstrans[TRANSITIONS];

/* This function, called from timing, finds us the next break.  It's
   been moved to a separate function because it was found desirable
   to use it in multiple situations.
   
   It returns the location of the end of the next break if there is one,
   or -1 if it was not found.  The location of the break is returned as
   the counter index into the rstrans array of the transition that ends
   the break, not as a time.  The same is true for the position input
   variable.
*/
long nextbreak(long position, long length)
{
    long counter;
    
    /* If they said to start at 0, they really meant 1.  Because we
       refer to counter-1, we can't start at 0 */
    position=position?position:position+1;

    for (counter=position; counter<TRANSITIONS; counter++)
    {
	if (rstrans[counter].time==-1)
	    break;
	
	if  (rstrans[counter].time-rstrans[counter-1].time > length
		&& rstrans[counter-1].state)
	{
	    /* Return the end of the found break */
	    return(counter);
	}
    }
    return -1; /* Didn't find any */
}

void dump(int signal, long baud, int wordsize, int parity, int verbose, int flags)
{
    unsigned char mask, prevstate, state, reversed=0;
    long counter, counter2, prevtime, totaltransitions;
    long endofbreak;
    unsigned int value;
    long shortest, numtrans;
    long glitches[TRANSITIONS/100], glitchcount;
    char afterbreak=1;

    /* This code is taken from the rs232 program, and should be
       put into a separate function */

    for (counter=0; counter<TRANSITIONS; counter++)
    {
	if (scanf("%d %ld %x\n", &transitions[counter].tick,
	    &transitions[counter].loop, &transitions[counter].value.byte)
	    != 3)
	    break;
	if (transitions[counter].tick==0)
	    break;
    }
    
    mask=(1<<signal)<<4;
    
    prevstate=transitions[0].value.byte&mask;
    totaltransitions=0;
    for (counter=0; counter<TRANSITIONS; counter++)
    {
	if (transitions[counter].tick==0)
	    break;

	if ((transitions[counter].value.byte&mask) != prevstate)
	{
	    prevstate=(transitions[counter].value.byte&mask);
	    rstrans[totaltransitions].time=transitions[counter].loop;
	    rstrans[totaltransitions].state=prevstate;
	    totaltransitions++;
	}
    }
    rstrans[totaltransitions].time=-1; /* mark the end */
    
    if (verbose)
	printf (" Total of %ld transitions on indicated line.\n", totaltransitions);
    
    if (totaltransitions < 4)
    {
	printf ("There does not seem to be any RS232 data here.\n");
	exit(1);
    }

    if (!flags&0x01) /* Reverse the data if indicated */
    {
	for (counter=0; counter<TRANSITIONS; counter++)
	{
	    if (rstrans[counter].time==-1)
		break;
	    
	    rstrans[counter].state=!rstrans[counter].state;
	}

	if (verbose==2) 
	    printf ("  Raw data reversed.\n");
    
    }
    
    shortest=1000000/baud;
    
    if (verbose==2)
    {
	printf ("  Bit length is %ld usec\n", shortest);
	printf ("  Word length is %d bits\n", wordsize);
	printf ("  Parity number is %d\n", parity);
    }
    
    endofbreak=nextbreak(0,shortest*(wordsize+(!!parity)+2));
    if (endofbreak==-1)
    {
	printf ("No break!\n");
	exit(1);
    }
    if (verbose==2)
	printf ("  End of first break at %ld usec\n", rstrans[endofbreak].time);
    
    glitchcount=0;
    for (counter=1; counter<TRANSITIONS; counter++)
    {
	if (rstrans[counter].time==-1)
	    break;
	
	if (rstrans[counter].time-rstrans[counter-1].time < shortest)
	{
	    if (rstrans[counter].time-rstrans[counter-1].time <
		    shortest-shortest/15)
	    {
		if (verbose==2)
		    printf ("  Glitch of size %ld at %ld usec using bitsize %ld usec\n",
			    rstrans[counter].time-rstrans[counter-1].time,
			    rstrans[counter].time, shortest);
		glitches[glitchcount]=counter;
		glitchcount++;
	    }
	}
    }
    
    /* Mark end of glitches. */
    glitches[glitchcount]=MAXLONG;
    
    /* Now we dump the data.  At the moment, it's interpreted in LSB
       first format and displayed in hexadecimal.  Incorrect parity is
       indicated with a trailing underscore.  Breaks are a carriage
       return, and glitches say so.
     */

    counter2=endofbreak;
    prevtime=rstrans[counter2].time;
    
    glitchcount=0;
    while (rstrans[counter2].time != -1 && counter2 != -1)
    {
	if (verbose==2)
	    printf ("(%ld)", prevtime);

	if (verbose && afterbreak)
	    printf ("[%ld] ", rstrans[counter2].time-rstrans[counter2-1].time);
	afterbreak=0;

	/* prevstate is used as a scratch variable to count the number of
	   high bits in the word.  Value notes the value using LSB format.
	 */
	prevstate=0;
	value=0;

	/* Get data from all the data bits in the word. */
	for (counter=0; counter<(wordsize+!!parity); counter++)
	{
	    /* Advance to the right place */
	    for (; rstrans[counter2].time < prevtime+shortest*(1.5+counter);
		    counter2++)
		state=!!rstrans[counter2].state;

	    prevstate += state;
	    
	    /* Don't count the parity bit */
	    if (counter != wordsize)
		value += state*(1<<counter);
	}
	if (rstrans[counter2].time == -1 || counter2==-1)
	    break;

	/* Now check to see if we've passed a glitch.  If we have, 
	   skip everything between here and the next break.
	 */
	if (glitches[glitchcount]<counter2)
	{
	    printf ("*\n");
	    afterbreak=1;
	    glitchcount++;
	    counter2=nextbreak(counter2, shortest*(wordsize+(!!parity)+2));
	    prevtime=rstrans[counter2].time;
	    continue;
	}

	/* Now advance into the stop bit, but save the state for parity */
	prevstate=state;
	for (; rstrans[counter2].time < prevtime+shortest*(1.5+(wordsize+!!parity));
		counter2++)
	    state=!!rstrans[counter2].state;

	/* Make sure there's a stop bit here */
	if (!state)
	{
	    printf ("** ");
	    counter2=nextbreak(counter2, shortest*(wordsize+(!!parity)+2));
	    prevtime=rstrans[counter2].time;
	    continue;
	}

	/* At this point, prevstate will have the total number of bits that
	   were high, and state will have the state of the last (parity)
	   bit.  From this, we can eliminate things.
	 */
	
	printf ("%x", value);
	
	/* Now print parity indication.  Parity is N O E M S, in that order */
	if  (parity == 0 ||
	    (parity==1 && prevstate%2==1) ||
	    (parity==2 && prevstate%2==0) ||
	    (parity==3 && prevstate==1) ||
	    (parity==4 && prevstate==0))
	    printf (" ");
	else
	    printf ("_");
	
	/* See if we've had a break */
	if (rstrans[counter2].time-prevtime>shortest*(wordsize*2+2))
	{
	    printf ("\n");
	    afterbreak=1;
	}

	prevtime=rstrans[counter2].time;
    }	
    
}

int main (int argc, char*argv[])
{
    int i, wordsize=8, parity=0, flags=0;
    int signal=0, verbose=0;
    long baud=9600;
    
    for (i=1; i<argc; i++)
    {
	if (strcmp(argv[i], "-s")==0)
	    signal=atoi(argv[++i]);
	else if (strcmp(argv[i], "-w")==0)
	{
	    wordsize=atoi(argv[++i]);
	}
	else if (strcmp(argv[i], "-b")==0)
	{
	    baud=atol(argv[++i]);
	}
	else if (strcmp(argv[i], "-r")==0)
	{
	    flags|=0x01;
	}
	else if (strcmp(argv[i], "-p")==0)
	{
	    i++;
	    switch (argv[i][0])
	    {
		case 'n':
		case 'N': parity=0;
			  break;
		case 'o':
		case 'O': parity=1;
			  break;
		case 'e':
		case 'E': parity=2;
			  break;
		case 'm':
		case 'M': parity=3;
			  break;
		case 's':
		case 'S': parity=4;
			  break;
		default:  printf ("Parity must be N, O, E, M, or S.\n");
			  exit(255);
	    }

	}
	else if (strcmp(argv[i], "-v")==0)
	    verbose=1;
	else if (strcmp(argv[i], "-d")==0)
	    verbose=2;
	else    
	{
	    printf ("Valid options are:\n\n"
		     "  -s n  Signal is (0: CTS 1: DSR 2: RI 3: DCD)\n"
		     "        Defaults to CTS\n"
		     "  -b n  Interpret as n bits per second (default 9600)\n"
		     "  -w n  Use word size of n bits (default 8)\n"
		     "  -r    Data is reversed\n"
		     "  -p x  Use parity x (Mark, Space, Odd, Even, or None)\n"
		     "        Defaults to None\n"
		     "  -v    Verbose output\n"
		     "  -d    Debugging output\n\n");
		
	    exit(255);
	}
    }
    
    if (verbose)
	printf (" Logic Analyzer Data Dumper version 0.1\n\n");
    
    dump(signal, baud, wordsize, parity, verbose, flags);

    return 0;
}
