/*
    Serial port logic analyzer suite
    Copyright 1999 Terran Melconian (terran@consistent.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program, in the file "LICENSE".  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
    Boston, MA  02111-1307  USA

*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/io.h>
#include <string.h>
#include <values.h>

#define TRANSITIONS 100000
#define MAXWORDSIZE 100

/* This is a set of "standadr" baud rates we snap to if it's close */
long baudrate[]={50, 75, 110, 150, 300, 600, 1200, 2400, 4800, 9600,
    19200, 38400, 57600, 115200, 0};

union MSR /* Modem Status Register */
{
    char byte;
    struct
    {
	int DCTS:1;
	int DDSR:1;
	int TERI:1;
	int DDCD:1;
	int CTS:1;
	int DSR:1;
	int RI:1;
	int DCD:1;
    } flags;
};

struct TRANSITION /* Recording signal level transitions */
{
    int tick; /* How many clock ticks in from beginning */
    long loop; /* and how many loops since that tick */
    union MSR value; /* Record the new values */
};

struct TRANSITION transitions[TRANSITIONS];

struct RSTRANS /* Modified version for use here */
{
    unsigned long time; /* time in microseconds */
    char state; /* New state, 1 or 0 */
};
struct RSTRANS rstrans[TRANSITIONS];

/* This function, called from timing, finds us the next break.  It's
   been moved to a separate function because it was found desirable
   to use it in multiple situations.
   
   It returns the location of the end of the next break if there is one,
   or -1 if it was not found.  The location of the break is returned as
   the counter index into the rstrans array of the transition that ends
   the break, not as a time.  The same is true for the position input
   variable.
*/
long nextbreak(long position, long length)
{
    long counter;
    
    /* If they said to start at 0, they really meant 1.  Because we
       refer to counter-1, we can't start at 0 */
    position=position?position:position+1;

    for (counter=position; counter<TRANSITIONS; counter++)
    {
	if (rstrans[counter].time==-1)
	    break;
	
	if  (rstrans[counter].time-rstrans[counter-1].time > length
		&& rstrans[counter-1].state)
	{
	    /* Return the end of the found break */
	    return(counter);
	}
    }
    return -1; /* Didn't find any */
}

void timing(int signal, int wordsize, int verbose, int mach)
{
    unsigned char mask, prevstate, state, reversed=0;
    unsigned int wordsizeerror[MAXWORDSIZE+1]; /* 5 to 9  or more if extended */
    unsigned int parityerror[4]={0,0,0,0}; /* O E M S */
    /* Odd parity means the total number of high(mark) bits, including the
       parity bit, is odd */
    long counter, counter2, prevtime, totaltransitions;
    long endofbreak;
    long shortest, baud, numtrans;
    long glitches[TRANSITIONS/100], glitchcount;
    int temp;
    char startwordsize=5;
    char stopwordsize=9;
    
    if (wordsize > 9)
    {
	stopwordsize=wordsize/10;
	wordsize%=10;
    }
    
    /* First we just read in the data */

    for (counter=0; counter<TRANSITIONS; counter++)
    {
	if (scanf("%d %ld %x\n", &transitions[counter].tick,
	    &transitions[counter].loop, &temp)
	    != 3)
	    break;
	transitions[counter].value.byte=temp;
	if (transitions[counter].tick==0)
	    break;
    }
    
    /* This will give us an AND mask for the signal line we're interested in. */
    mask=(1<<signal)<<4;
    
    /* I think the first thing we actually want to do is make a note of the
       transitions of the pin we're interested in and store them somewhere
       separately.  This will make life substantially easier than having to
       go through and pick out only the ones we want every time.  This list
       will start at the first transition of the indicated signal, not at
       the beginning of the data file.
     */
    
    prevstate=transitions[0].value.byte&mask;
    totaltransitions=0;
    for (counter=0; counter<TRANSITIONS; counter++)
    {
	if (transitions[counter].tick==0)
	    break;

	if ((transitions[counter].value.byte&mask) != prevstate)
	{
	    prevstate=(transitions[counter].value.byte&mask);
	    rstrans[totaltransitions].time=transitions[counter].loop;
	    rstrans[totaltransitions].state=prevstate;
	    totaltransitions++;
	}
    }
    rstrans[totaltransitions].time=-1; /* mark the end */
    
    if (verbose)
	printf (" Total of %ld transitions on indicated line.\n", totaltransitions);
    
    if (totaltransitions < 4)
    {
	printf ("There does not seem to be any RS232 data here.\n");
	exit(1);
    }
    
    /* Now we look for the baud rate.  At first I just took the shortest
       time between transitions, but I'm having problems with glitches
       in the data.  Therefore, I'm now taking the shortest time between
       transitions for which at least 2% of the segments are this value.
       
       This make the code a lot more complicated. Enh.  Such is life.
    */

    /* Here, shortest keeps track of the shortest transition we have
       found.  numtrans keeps track of how many we've found within ten
       percent of that length.  If this is fewer than 2% of the total
       number of transitions, we punt and rerun, ignoring those.  prevtime
       is used to keep track of the length below which we're ignoring.
    */
    prevtime=0;
    while(1)
    {
	numtrans=0;
	shortest=MAXLONG;
	counter2=0;
	for (counter=1; counter<TRANSITIONS; counter++)
	{
	    if (rstrans[counter].time==-1)
		break;

	    if (verbose==2 && !prevtime)
		printf ("  Bit at %ld usec of length %ld\n",
			rstrans[counter].time,
			rstrans[counter].time-rstrans[counter2].time);


	    if (rstrans[counter].time-rstrans[counter2].time < shortest
		&&rstrans[counter].time-rstrans[counter2].time > prevtime)
		shortest=rstrans[counter].time-rstrans[counter2].time;
	    counter2=counter;
	}
	
	/* Now we have to do it all again to find out how *many* there
	   were within some percentage of this size.  Sigh.
	   
	   We also make a note of every one that's *smaller* than this
	   as a glitch, for later use in skipping over it.
	 */
	glitchcount=0;
	for (counter=1; counter<TRANSITIONS; counter++)
	{
	    if (rstrans[counter].time==-1)
		break;
	    
	    if (rstrans[counter].time-rstrans[counter-1].time < shortest)
	    {
		if (rstrans[counter].time-rstrans[counter-1].time <
			shortest-shortest/15)
		{
		    if (verbose==2)
			printf ("  Glitch of size %ld at %ld usec using bitsize %ld usec\n",
				rstrans[counter].time-rstrans[counter-1].time,
				rstrans[counter].time, shortest);
		    glitches[glitchcount]=counter;
		    glitchcount++;
		    if (glitchcount >= TRANSITIONS/100)
		    {
			fprintf(stderr, "This data is fscked.\n");
			exit(1);
		    }
		}
	    }
	    else if (rstrans[counter].time-rstrans[counter-1].time < 
		shortest+shortest/10)
		numtrans++;
	}
	
	/* Mark end of glitches. */
	glitches[glitchcount]=MAXLONG;

	if (verbose) 
	{
	    printf (" Shortest transition %ld microseconds.  ", shortest);
	    printf (" %ld segments within 10%% of this.\n", numtrans);
	}
	
	if (numtrans >= totaltransitions/50)
	    break;
	else
	{
	    /* we need to make sure we round *up* here, not down, which
	       is why the arithmetic is screwy.  For example, saying something
	       like prevtime=shortest+shortest/10 would leave us in an
	       infinite loop if we were looking at a shortest time of 3 usec
	     */
	    prevtime=2*shortest-9*shortest/10;
	    continue;
	}
    }
    
    /* Now we actually average the times for the bits within 10% (or
       whatever percentage) of the shortest one.  We'll average only
       as many as we can without overflowing the variables; that should
       certainly be enough to give us a good figure anyway
       
       prevtime is used as the totaler, in a gratuitious and unmaintainable
       case of using a random scratch variable.
     */
    counter2=0;
    prevtime=0;
    numtrans=0;
    for (counter=1; counter<TRANSITIONS; counter++)
    {
	if (rstrans[counter].time==-1)
	    break;

	if (rstrans[counter].time-rstrans[counter2].time > shortest
	    &&rstrans[counter].time-rstrans[counter2].time < 
	    shortest+shortest/10)
	{
	    numtrans++;
	    prevtime+=rstrans[counter].time-rstrans[counter2].time;
	    
	    /* avoid overflowing */
	    if (prevtime > MAXLONG-2*shortest)
		break;
	}
	counter2=counter;
    }
	
    // This happens when we looked at data that had segments at two
    // radically different baud rates
    if (!prevtime || !numtrans)
    {
	fprintf (stderr, "This data is fscked.\n");
	exit(1);
    }
    baud=1000000/(prevtime/numtrans);

    if (verbose)
    {
	printf (" Calculated baud rate is %ld bits per second.\n", baud);
    }
    
    for (counter=0; baudrate[counter]; counter++)
    {
	if (baud*1.04 > baudrate[counter] && baud*0.95 < baudrate[counter])
	{
	    baud=baudrate[counter];
	    printf ("Baud rate is %ld bits per second.\n", baud);
	    
	    /* Now reset the shortest length */
	    shortest=1000000/baud;

	    break;
	}
    }
    if (!baudrate[counter]) /* if we didn't find a match */
    {
	printf ("Baud rate of %ld bps does not look standard.  "
		"Not performing any rounding.\n", baud);
    }
    
    /* Now we start trying to find the number of bits in various places,
       and the parities, etc.  This part is less certain, but we can have
       a go at it, and we rather need this in order to have done anything
       nontrivial.  We also allow for the possibility that the signals
       are reversed (high is low and low is high).  If so, we say so.
       
       The first thing we do is look for breaks, but before that we see
       if we need to reverse the data.
     */
    
reversestart:
    
    if (reversed) /* Reverse the data if this is our second try */
    {
	for (counter=0; counter<TRANSITIONS; counter++)
	{
	    if (rstrans[counter].time==-1)
		break;
	    
	    rstrans[counter].state=!rstrans[counter].state;
	}
    }
    
    /* Now look for a break (a mark segment longer than one byte) */
    
    endofbreak=nextbreak(0,shortest*(stopwordsize+2));
    if (endofbreak == -1) /* If there aren't any */
    {
	if (reversed)
	{
	    printf ("Could not find any breaks.  Try a longer sample.\n");
	    exit(1);
	}
	reversed=1;
	if (verbose)
	    printf (" Could not find break with default polarity.  "
		    "Trying reverse.\n");
	goto reversestart;
    }
    
    if (verbose)
	printf (" First break ends at %ld usec.\n", rstrans[endofbreak].time);
    
    /* Make sure we don't have breaks of both polarities

       This doesn't call the nextbreak() function, because it does not
       have the capability to deal with reversed polarities unless the
       actual data is flipped
     */
    for (counter=1; counter<TRANSITIONS && rstrans[counter].time!=-1; counter++)
    {
	if  (rstrans[counter].time-rstrans[counter-1].time > shortest*11
		&& rstrans[counter-1].state==0)
	{
	    if (verbose)
		printf (" Found \"breaks\" with both polarities.\n");
	    printf ("This doesn't look like RS232.  Sorry.\n");
	    exit(1);
	}
    }
    
    /* Now output the polarity.  Our input polarity is actually reversed
       from standard RS232 - the status lines read 0 when the voltage is neg
       and 1 when it's positive, but the data lines are the opposite.  Thus,
       if we had to reverse, it means it's actually correct.  Confused yet?
     */
    printf ("Polarity is %s\n", reversed?"correct":"reversed");
    
    /* Now we start counting the bits.  The first task is to determine
       the sum of the data and parity bits.  We simply go to the first
       break, and try stepping through the available data with all
       possibilities.  We rule out those which demonstrably do not fit,
       and see what's left.  If we haven't ruled out all but one by then,
       we quit, as there's nothing more we can do.
       
       If we come to a location where there is a known glitch, we skip
       all information until the next break.
     */

    for (counter=startwordsize; counter<=stopwordsize; counter++)
    {
	glitchcount=0;
	wordsizeerror[counter]=0;
	counter2=endofbreak;
	prevtime=rstrans[counter2].time;
	
	while (rstrans[counter2].time != -1 && counter2 != -1)
	{
	    for (; rstrans[counter2].time < prevtime+shortest*(counter+1.5);
		    counter2++)
		state=rstrans[counter2].state;
	    
	    /* Now check to see if we've passed a glitch.  If we have, 
	       skip everything between here and the next break.
	     */
	    if (glitches[glitchcount]<counter2)
	    {
		glitchcount++;
		counter2=nextbreak(counter2, shortest*(stopwordsize+2));
		prevtime=rstrans[counter2].time;
		continue;
	    }

	     /* If we're not in a stop bit, rule out this one */
	    if (!state && rstrans[counter2].time != -1)
	    {
		wordsizeerror[counter]++;
		if (verbose==2)
		    printf ("  Bad word for size %ld in word starting at %ld usec\n",
			    counter, prevtime);

		counter2=nextbreak(counter2,shortest*(stopwordsize+2));
	    }
	    
	    /* Now advance to the next word.  We'll already be on the
	       next transition after the stop bit, from before.
	     */
	    prevtime=rstrans[counter2].time;
	}	
    }
    
    /* If verbose, dump all our error counts */
    /* Also count up how many perfect sizes there were */
    /* prevtime is used to note the size with the errors, for use
       in selecting the best word size later */
    counter2=0;
    prevtime=-1;
    for (counter=startwordsize; counter<=stopwordsize; counter++)
    {
	if (prevtime==-1 || wordsizeerror[counter] < wordsizeerror[prevtime])
	    prevtime=counter;

	counter2+=!wordsizeerror[counter];

	if (verbose)
	    printf (" Word size %ld:  %u errors\n", counter,
		    wordsizeerror[counter]);
    }
    
    /* counter2 contains the number of perfect sizes */
    if (counter2 > 1 && !wordsize)
    {
	printf ("Possible word sizes: ");
	for (counter=startwordsize; counter<=stopwordsize; counter++)
	    if (!wordsizeerror[counter])
		printf (" %ld", counter);
	printf (" including any parity bits.\n"
		"Try forcing a word size with the -w option.\n");
	exit(0);
    }
    if (wordsize)
    {
	printf ("Selected wordsize of %d bits has %u errors.\n",
		wordsize, wordsizeerror[wordsize]);
    }
    else if (counter2==0)
    {
	printf ("No perfect word sizes.\n"
		"Best is %ld bits (including any parity), which had %u errors.\n",
		prevtime,wordsizeerror[prevtime]);
	wordsize=prevtime;
    }
    else
    {
	printf ("Word size is %ld bits, including any parity bits.\n",
		prevtime);
	wordsize=prevtime;
    }

    /* I believe that the next thing to do is to determine the number of
       stop bits, because this won't be very hard.  Of course, data with
       *any* number of stop bits can be successfully interpreted using 1
       stop bit, so it's not clear how much we really *care* about how
       many stop bits there are.  And of course, any data with two stop
       bits won't have its word length successfully detected, so it will
       have to be forced.
     */
    
    /* This is being skipped for the time being due to its uselessness */
    
    /* Now we need to check to see what the possibilities for parity
       are.  We simply step through the words, and rule out those
       that are wrong, just as we did for word length.
     */
    counter2=endofbreak;
    prevtime=rstrans[counter2].time;
    
    glitchcount=0;
    while (rstrans[counter2].time != -1 && counter2 != -1)
    {
	/* prevstate is used as a scratch variable to count the number of
	   high bits in the word
	 */
	prevstate=0;

	/* Get data from all the data bits in the word. */
	for (counter=0; counter<wordsize; counter++)
	{
	    /* Advance to the right place */
	    for (; rstrans[counter2].time < prevtime+shortest*(1.5+counter);
		    counter2++)
		state=!!rstrans[counter2].state;

	    prevstate += state;
	}
	if (rstrans[counter2].time == -1)
	    break;

	/* Now check to see if we've passed a glitch.  If we have, 
	   skip everything between here and the next break.
	 */
	if (glitches[glitchcount]<counter2)
	{
	    glitchcount++;
	    counter2=nextbreak(counter2, shortest*(stopwordsize+2));
	    prevtime=rstrans[counter2].time;
	    continue;
	}

	/* At this point, prevstate will have the total number of bits that
	   were high, and state will have the state of the last (parity)
	   bit.  From this, we can eliminate things.
	 */
	
	if (state) /* remove mark or space possibilities */
	{
	    parityerror[3]++;
	}
	else
	{
	    parityerror[2]++;
	}
	    
	
	if (prevstate % 2) /* remove odd or even possibilities */
	{
	    if (verbose==2)
		printf ("  Even parity error from word starting at %ld usec\n",
			prevtime);

	    parityerror[1]++;
	}
	else
	{
	    if (verbose==2)
		printf ("  Odd parity error from word starting at %ld usec\n",
			prevtime);

	    parityerror[0]++;
	}

	/* Now advance into the stop bit and set up for next word*/
	for (; rstrans[counter2].time < prevtime+shortest*(1.5+wordsize);
		counter2++)
	    state=!!rstrans[counter2].state;

	prevtime=rstrans[counter2].time;
    }	
    
    /* This one is easier to do as multiple printfs */
    if (verbose)
    {
	printf (" Parity errors for %do: %d\n", wordsize-1, parityerror[0]);
	printf (" Parity errors for %de: %d\n", wordsize-1, parityerror[1]);
	printf (" Parity errors for %dm: %d\n", wordsize-1, parityerror[2]);
	printf (" Parity errors for %ds: %d\n", wordsize-1, parityerror[3]);
    }
    
    /* Hm.  Just print the ones which were perfect; if we want the
       fuzzy data, we can use the verbose option. */
    if (!parityerror[0] || !parityerror[1] || !parityerror[2] || 
	    !parityerror[3] || wordsize<9 || stopwordsize > 9)
    {
	printf ("Possible parities: ");
	if (wordsize != 9 || stopwordsize > 9)
	    printf ("%dn ", wordsize);
	for (counter=0; counter<4; counter++)
	    if (!parityerror[counter])
		printf ("%d%c ",wordsize-1,"oems"[counter]);
	printf ("\n");
    }
    else
	printf ("No error-free parities detected.\n");
}

int main (int argc, char*argv[])
{
    int i, wordsize=0;
    int signal=0, verbose=0, mach=0;
    
    for (i=1; i<argc; i++)
    {
	if (strcmp(argv[i], "-s")==0)
	    signal=atoi(argv[++i]);
	else if (strcmp(argv[i], "-w")==0)
	{
	    wordsize+=atoi(argv[++i]);
	}
	else if (strcmp(argv[i], "-x")==0)
	{
	    wordsize+=atoi(argv[++i])*10;
	    if (wordsize/10 > MAXWORDSIZE)
	    {
		printf ("Word size of %d is greater than compiled-in max of %d.\n",
			wordsize/10, MAXWORDSIZE);
		exit(255);
	    }
	}
	else if (strcmp(argv[i], "-v")==0)
	    verbose=1;
	else if (strcmp(argv[i], "-d")==0)
	    verbose=2;
	else    
	{
	    printf ("Valid options are:\n\n"
		     "  -s n  Signal is (0: CTS 1: DSR 2: RI 3: DCD)\n"
		     "        Defaults to CTS\n"
		     "  -w n  Force word size to n bits (5 to 9), including any parity\n"
		     "  -x n  Use extended word size range (1 to n)\n"
		     "  -m    Machine-readable output\n"
		     "  -v    Verbose output\n"
		     "  -d    Debugging output\n\n");
		
	    exit(255);
	}
    }
    
    if (verbose)
	printf (" Logic Analyzer RS232 Timer version 0.2\n\n");
    
    timing(signal, wordsize, verbose, mach);

    return 0;
}
