/*
    Serial port logic analyzer suite
    Copyright 1999,2004 Terran Melconian (terran@consistent.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program, in the file "LICENSE".  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
    Boston, MA  02111-1307  USA

*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/times.h>
#include <sys/time.h>
#include <string.h>
#include <sched.h>
#include <sys/io.h>

/* TODO: remove hardcoded limit */
#define TRANSITIONS 1000000

int usehighpriority=1;

union MSR /* Modem Status Register */
{
    char byte;
    struct
    {
	int DCTS:1;
	int DDSR:1;
	int TERI:1;
	int DDCD:1;
	int CTS:1;
	int DSR:1;
	int RI:1;
	int DCD:1;
    } flags;
};

struct TRANSITION /* Recording signal level transitions */
{
    struct timeval tv; /* Time at which the transition occurred */
    union MSR value; /* Record the new values */
};

int highpriority(void)
{
    struct sched_param p;

    if (usehighpriority==0)
	return 0;

    p.sched_priority=1;
    return sched_setscheduler(0, SCHED_FIFO, &p);
}

int lowpriority(void)
{
    struct sched_param p;

    p.sched_priority=0;
    return sched_setscheduler(0, SCHED_OTHER, &p);
}

struct TRANSITION transitions[TRANSITIONS];

void sample(unsigned baseaddr, int duration)
{
    union MSR msr, prior;
    struct timeval tv, starttv;
    long i, count;

    if (ioperm(baseaddr,8,1))
    {
	fprintf(stderr, "You must have root permissions.\n");
	exit(1);
    }

    /* Initializations before the timing starts */
    /* Make them different so it samples on the first run */
    prior.byte=1;
    
    highpriority();

    gettimeofday(&starttv, 0);

    count=0;
    i=0;
    while(1)
    {
	msr.byte=inb(baseaddr+6);
	msr.byte&=0xf0;
	count++;

	/* If something changed, get the time and store the data */
	if (msr.byte != prior.byte)
	{
	    prior.byte = msr.byte;
	    gettimeofday(&transitions[i].tv, 0);
	    transitions[i].value.byte=msr.byte;

	    /* See if we're done */
	    if (transitions[i].tv.tv_sec - starttv.tv_sec >= duration)
		break;
	    count=0;

	    if (++i==TRANSITIONS) // Don't overflow fixed memory
	    {
		i--;
		break;
	    }
	}

	/* However, it's possible NOTHING is changing, and we still need to
	 * exit after the specified amount of time.  */
	if (count > 10000)
	{
	    gettimeofday(&tv, 0);
	    if (tv.tv_sec - starttv.tv_sec >= duration)
		break;
	    count=0;
	}
	
    }

    /* Note end of data */
    transitions[i].tv.tv_sec=-1;;
    
    lowpriority();
}

void process(void)
{
    int i;
    long basetime = transitions[0].tv.tv_sec * 1000000 + transitions[0].tv.tv_usec;
    for (i=0; transitions[i].tv.tv_sec != -1; i++)
    {
	long transitiontime = transitions[i].tv.tv_sec * 1000000 +
	    transitions[i].tv.tv_usec;
	
	/* First field is no longer used for timing data; it's just used
	 * to indicate the end of file by being 0. */
	printf ("%ld %ld %x\n", (transitiontime-basetime)/10000+1, 
		transitiontime-basetime, transitions[i].value.byte);
    }
    printf ("%d %d %d\n", 0, 0, 0);
}

int main (int argc, char*argv[])
{
    int runtime=2, i;
    int address=0x3f8;
    
    for (i=1; i<argc; i++)
    {
	if (strcmp(argv[i], "-t")==0)
	    runtime=atoi(argv[++i]);
	else if (strcmp(argv[i], "-p")==0)
	    usehighpriority=0;
	else if (strcmp(argv[i], "-a")==0)
	{
	    i++;
	    sscanf(argv[i], "%i", &address);
	}
	else
	{
	    printf ("Valid options are:\n\n"
		     "  -p     Do NOT use high-priority scheduler\n"
		     "  -a n   address of serial port\n"
		     "  -t n   run for n seconds\n");
	    exit(255);
	}
    }
	
    sample(address, runtime);
    
    process();

    return 0;
}
